//------------------------------------------------------------------------
// Morph.java -- a simple engine for a graphical object world with
//               hierarchical struturing and layout management
//
// Andy Collins - CSE 590D - Spring 1997 - University of Washington
//------------------------------------------------------------------------
package application;

import java.awt.*;
import java.util.*;

class MorphPanel extends Panel
{
   private Vector morphs;
   private Morph selection;
   private int maxZValue;
   private int xMouseLoc;
   private int yMouseLoc;

   public int TopZSpot ()
   {
      ++maxZValue;
      return maxZValue;
   }

   private int GetIntersectAreaWith (int x, int y, int width, int height)
   {
      Rectangle rect = new Rectangle (x, y, width, height);
      Enumeration e = morphs.elements ();
      int area = 0;

      while (e.hasMoreElements ())
         {
         Morph m = (Morph)e.nextElement ();
         Rectangle r = new Rectangle (m.GetGlobalX (), m.GetGlobalY (),
                                      m.GetHeight (), m.GetWidth ());
         Rectangle i = r.intersection (rect);
         if (i.width > 0 && i.height > 0)
            area += 1000 * i.width * i.height;
         else
            area += i.width * i.height;
         }

      return area;
   }

   public Point FindEmptySpot (int width, int height)
   {
      int xrange = size().width - width;
      int yrange = size().height - height;
      if (xrange < 0)
         xrange = 0;
      if (yrange < 0)
         yrange = 0;

      int bestX = xrange/2;
      int bestY = yrange/2;
      int bestEval = GetIntersectAreaWith (bestX, bestY, width, height);

      for (int i=0; /*bestEval>0 &&*/ i<10; ++i)
         {
         int x = (int) (Math.random () * xrange);         
         int y = (int) (Math.random () * yrange);         
         int eval = GetIntersectAreaWith (x, y, width, height);
         if (eval < bestEval)
            {
            bestX = x;
            bestY = y;
            bestEval = eval;
            }
         }

      return new Point (bestX, bestY);
   }

   void RemoveAllMorphs ()
   {
      morphs.removeAllElements ();
   }

   MorphPanel ()
   {
      morphs = new Vector ();
      selection = null;
      maxZValue = 0;
      xMouseLoc = 0;
      yMouseLoc = 0;
   }

   void AddMorph (Morph m)
   {
      m.Layout ();
      morphs.addElement (m);
      m.SetPanel (this);
      MorphZChanged (m);
   }

   public void MorphZChanged (Morph m)
   {
      int newz = m.GetZ ();
      if (newz > maxZValue)
         maxZValue = newz;
      
      if (!morphs.contains (m) || morphs.size () < 2)
         {
         repaint ();
         return;
         }

      morphs.removeElement (m);

      for (int i=0; i<morphs.size (); ++i)
         {
         Morph test = (Morph) morphs.elementAt (i);
         int testZ = test.GetZ ();
         if (testZ > newz)
            {
            morphs.insertElementAt (m, i);
            repaint ();
            return;
            }
         }

      morphs.addElement (m);
      repaint ();
   }

   void RemoveMorph (Morph m)
   {
      morphs.removeElement (m);
      m.SetPanel (null);
      repaint ();
   }

   public void reshape (int x, int y, int width, int height)
   {
      super.reshape (x, y, width, height);
      makeDoubleBuffer (width, height);
   }

   protected Dimension offscreenSize;
   protected Image offscreenImage;
   protected Graphics offscreenGC = null;

   protected void makeDoubleBuffer (int width, int height)
   {
      if ((width > 0) && (height > 0))
         {
         offscreenSize = new Dimension (width, height);
         offscreenImage = createImage (width, height);
         offscreenGC = offscreenImage.getGraphics ();
         offscreenGC.setColor (new Color (240, 240, 240));
         offscreenGC.fillRect (0, 0, width, height);
         }
      else
         offscreenGC = null;
   }

   public void update (Graphics g)
   {
      paint (g);
   }

   public void ForceUpdate ()
   {
      Graphics g = getGraphics ();
      paint (g);
   }

   public void paint (Graphics g)
   {
      if (offscreenGC == null)
         makeDoubleBuffer (size ().width, size ().height);
      if (offscreenGC == null)
         return;

      Color background = Morph.defaultBgColor;
//      Color normal     = new Color (0,0,0);
//      Color selected   = new Color (128,0,0);

      offscreenGC.setColor (background);
      offscreenGC.fillRect (0,0,size().width,size().height);

      // paint the contents on offscreenGC
      Enumeration e = morphs.elements ();
      while (e.hasMoreElements ())
         {
         Morph m = (Morph) e.nextElement ();
         m.Paint (offscreenGC, 0, 0, background, false);
         }

      if (selection != null)
         selection.Paint (offscreenGC, 0, 0, background, true);

      g.drawImage (offscreenImage, 0, 0, this);
   }

   // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   // event and focus management

   Boolean haveFocus = Boolean.FALSE;
   public boolean gotFocus (Event evt, Object what)
   {
      synchronized (haveFocus)
         {
         haveFocus = Boolean.TRUE;
         }
      return (true);
   }
   public boolean lostFocus (Event evt, Object what)
   {
      synchronized (haveFocus)
         {
         haveFocus = Boolean.FALSE;
         }
      return (true);
   }

   private int dragXOff = 0;
   private int dragYOff = 0;

   private boolean uiEvent (Event evt, int x, int y, int key)
   {

/***************************************
      boolean needToGetFocus;
      synchronized (haveFocus)
         {
         needToGetFocus = !haveFocus.booleanValue ();
         }
      if (needToGetFocus)
         requestFocus ();
 ***************************************/

      if ((evt.id == Event.MOUSE_DRAG) && (selection != null) &&
          (x > 0) && (y > 0) && (x < size().width) && (y < size().height))
         {
//         selection.MoveTo (selection.GetX () + x - xMouseLoc,
//                           selection.GetY () + y - yMouseLoc);
         selection.MoveTo (dragXOff + x, dragYOff + y);
         repaint ();
         }

      if (evt.id == Event.MOUSE_DOWN)
         {
         for (int i = morphs.size ()-1; i >= 0; --i)
            if ((selection == null) && ((Morph) morphs.elementAt (i)).IsOver (xMouseLoc, yMouseLoc))
               {
               selection = (Morph) morphs.elementAt (i);
               morphs.removeElement (selection);
               dragXOff = selection.GetX () - x;
               dragYOff = selection.GetY () - y;
               
               ++maxZValue;
               selection.SetZ (maxZValue);
               }
         repaint ();
         }

      if ((evt.id == Event.MOUSE_MOVE) || (evt.id == Event.MOUSE_UP))
         {
         if (selection != null)
            {
            morphs.addElement (selection);
            Morph t = selection;
            selection = null;
            MorphZChanged (t);
            }
         }

      if ((evt.id == Event.MOUSE_MOVE) || (evt.id == Event.MOUSE_DRAG))
         {
         xMouseLoc = x;
         yMouseLoc = y;
         }

      return (true);
   }

   public boolean mouseMove (Event evt, int x, int y) 
   { 
      return (uiEvent (evt, x, y, -1)); 
   }
   public boolean mouseDrag (Event evt, int x, int y)
   { 
      return (uiEvent (evt, x, y, -1)); 
   }
   public boolean mouseUp   (Event evt, int x, int y)
   { 
      return (uiEvent (evt, x, y, -1)); 
   }
   public boolean mouseDown (Event evt, int x, int y)
   { 
      return (uiEvent (evt, x, y, -1)); 
   }
}

class Morph extends Object
{
   private MorphPanel panel;

   protected Vector morphs;
   protected int baseHeight;
   protected int baseWidth;
   private Morph parent;
   protected Color fgColor;
   protected Color bgColor;
   protected static Color defaultFgColor = new Color (0,0,0);
   protected static Color defaultBgColor = new Color (240, 240, 240);

   static int shadowOffset = 3;
   static Color shadowColor = new Color(200, 200, 200);

   public boolean solid;
   public boolean outline;

   public Morph (int xv, int yv, int zv, int w, int h, Color bg, Color fg)
   {
      panel = null;

      parent = null;
      morphs = new Vector ();
      x = xv;
      y = yv;
      z = zv;
      baseHeight = h;
      baseWidth = w;
      Layout ();

      solid = true;
      outline = false;

      bgColor = bg;
      fgColor = fg;

      
   }

   public Morph ()
   {
      this (0, 0, 0, 10, 10, defaultFgColor, defaultFgColor);
   }

   public void SetPanel (MorphPanel p)
   {
      panel = p;
   }

   public MorphPanel GetPanel ()
   {
      if (panel != null)
         return panel;
      else if (parent != null)
         return parent.GetPanel ();
      else
         return null;
   }

   public void SetFgColor (Color c)
   {
      fgColor = c;
   }

   public void SetBgColor (Color c)
   {
      bgColor = c;
   }

   public void SetHeight (int h)
   {
      baseHeight = h;
      Layout ();
   }
   public void SetWidth (int w)
   {
      baseWidth = w;
      Layout ();
   }

   public void AddMorph (Morph m)
   {
      morphs.addElement (m);
      m.parent = this;
      Layout ();
   }

   public void RemoveMorph (Morph m)
   {
      morphs.removeElement (m);
      Layout ();
   }

   private int x;
   private int y;
   private int z;

   public void SetZ (int zv)
   {
      z = zv;
      
   }
   public void MoveTo (int xv, int yv)
   {
      x = xv;
      y = yv;
   }
   public int GetX ()
   {
      return x;
   }
   public int GetY ()
   {
      return y;
   }
   public int GetZ ()
   {
      return z;
   }

   public int GetGlobalX ()
   {
      if (parent != null)
         return (GetX () + parent.GetGlobalX ());
      else
         return GetX ();
   }
   public int GetGlobalY ()
   {
      if (parent != null)
         return (GetY () + parent.GetGlobalY ());
      else
         return GetY ();
   }

   public boolean IsOver (int xv, int yv)
   {
      boolean retval = false;

      if (!solid)
         retval = false;
      else
         retval = ((xv > x) && (yv > y) && (xv < x + width) && (yv < y + height));

      Enumeration e = morphs.elements ();
      while (e.hasMoreElements ())
         if ( ((Morph) e.nextElement ()).IsOver (xv-x, yv-y) )
            retval = true;

      return retval;
   }

   protected int width;
   protected int height;

   public void Layout ()
   {
      Enumeration e = morphs.elements ();
      while (e.hasMoreElements ())
         ((Morph) e.nextElement ()).Layout ();

      width = baseWidth;
      height = baseHeight;
   }
   public int GetHeight ()
   {
      return height;
   }
   public int GetWidth ()
   {
      return width;
   }

   public void Paint (Graphics g, int xoff, int yoff, Color c, boolean selected)
   {
      Enumeration e;

      if (selected)
         {
         PaintBase (g, xoff+shadowOffset, yoff+shadowOffset, shadowColor, shadowColor);
         e = morphs.elements ();
         while (e.hasMoreElements ())
            ((Morph) e.nextElement ()).Paint (g, xoff + x, yoff + y, bgColor, true);
         }

      PaintBase (g, xoff, yoff, bgColor, fgColor); 
      e = morphs.elements ();
      while (e.hasMoreElements ())
         ((Morph) e.nextElement ()).Paint (g, xoff + x, yoff + y, c, false);
   }

   public void PaintBase (Graphics g, int xoff, int yoff, Color bgColor, Color fgColor)
   {
      if (solid)
         {
         g.setColor (bgColor);
         g.fillRect (xoff + x, yoff + y, width, height);
         }

      if (outline)
         {
         g.setColor (fgColor);
         g.drawRect (xoff + x, yoff + y, width, height);
         }
   }
}

class LabelMorph extends Morph
{
   private boolean textSolid;

   private String label;
   public LabelMorph (String l)
   {
      label = l;
      SetHeight (15);
      SetWidth (50);
      textSolid = false;
   }

   public void PaintBase (Graphics g, int xoff, int yoff, Color bgColor, Color fgColor)
   {
      super.PaintBase (g, xoff, yoff, bgColor, fgColor);

      Font font = new Font ("TimesRoman", Font.PLAIN, 12);
      FontMetrics fm = g.getFontMetrics (font);

      if (textSolid && !solid)
         {
         g.setColor (bgColor);
         g.fillRect (xoff + GetX (), yoff + GetY (), fm.stringWidth (label), height);
         }

      g.setColor (fgColor);
      g.drawString (label, xoff + GetX (), yoff + GetY() + GetHeight () - 2);
   }
}

class ColumnMorph extends Morph
{
   private int inset;
   public void SetInset (int i)
   {
      inset = i;
   }

   public void Layout ()
   {
      super.Layout ();

      height = 0;
      width = baseWidth;
      Enumeration e;

      e = morphs.elements ();
      while (e.hasMoreElements ())
         {
         Morph m = (Morph) e.nextElement ();
         height += m.GetHeight ();
         if (m.GetWidth () > width)
            width = m.GetWidth ();
         }

      if (height < baseHeight)
         height = baseHeight;

      height += 2*inset;
      width += 2*inset;

      int curY = inset;
      e = morphs.elements ();
      while (e.hasMoreElements ())
         {
         Morph m = (Morph) e.nextElement ();
         m.MoveTo (inset, curY);
         curY += m.GetHeight (); 
         }
   }

   public void PaintBase (Graphics g, int xoff, int yoff, Color bgColor, Color fgColor)
   {
      super.PaintBase (g, xoff, yoff, bgColor, fgColor);
//      g.drawRect (xoff + x, yoff + y, width, height);
   }

}

