package application;

import network.*;
import link.*;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.net.*;

public class Nebraska extends Frame implements PacketReceiver
{
   public static void main (String[] args)
   {
      Link theLink = null;
      int port = 0;
      try
         {
         port = Integer.parseInt (args[1]);
         }
      catch (Exception e)
         {
         System.out.println ("error setting up");
         }

      if (args.length == 2)
         {
         theLink = new UDPLink ("link from " + args[0], 1000, port, 0.9, 10.0);
         }
      else if (args.length == 3)
         {
         theLink = new UDPLink ("link from " + args[0], 1000, args[2], port, 0.9, 10.0);
         }
      else
         {
         System.out.println ("USAGE: java Nebraska local_host port remote_host");
         System.exit (1);
         }

      try
         {
         Nebraska program = new Nebraska ();


         RoutingNetworkEdge e = new RoutingNetworkEdge ();
         theLink.AddNode (e);
         e.AddLink (theLink);

         NebraskaMorph.network = e;
         NebraskaMorph.address = new Address (args[0]);
         NebraskaMorph.network.AddReceiver (program, NebraskaMorph.address);

         program.Go ();
         }
      catch (Exception e)
         {
         System.out.println ("error setting up");
         }
   }

   private Nebraska ()
   {
      super ("Nebraska");
   }

   public void PacketArrived (NetworkPacket p)
   {
      byte[] data = p.data;
      int strlen = (data[0] << 8) + data[1];

      String msg = new String (data, 0, 2, strlen);

      NebraskaMorph.ProcessUpdate (msg);
   }

   private MorphPanel morphPanel;
   private TextField colorEntryRed;
   private TextField colorEntryGreen;
   private TextField colorEntryBlue;
   private TextField xSizeEntry;
   private TextField ySizeEntry;

   private Button newMorph;

   private void Go ()
   {
      SetupWindow ();
      resize (500, 400);
      show ();
   }

   public boolean handleEvent (Event e)
   {
      if (e.id == Event.WINDOW_DESTROY)
         {
         System.exit (0);
         }
      else if (e.target == newMorph && e.id == Event.ACTION_EVENT)
         {
         try
            {
            int red = Integer.parseInt (colorEntryRed.getText ());
            int green = Integer.parseInt (colorEntryGreen.getText ());
            int blue = Integer.parseInt (colorEntryBlue.getText ());

            int xsize = Integer.parseInt (xSizeEntry.getText ());
            int ysize = Integer.parseInt (ySizeEntry.getText ());

            Color color = new Color (red, green, blue);

            Point p = morphPanel.FindEmptySpot (xsize, ysize);

            NebraskaMorph m = new NebraskaMorph (p.x, p.y, morphPanel.TopZSpot (), 
                                                 xsize, ysize, color, Color.black);

            morphPanel.AddMorph (m);
//            m.MoveTo (p.x, p.y);
            }
         catch (Exception ex)
            {
            System.out.println ("Exception " + ex);
            }
         }
      return super.handleEvent (e);
   }

   private void SetupWindow ()
   {
      setLayout (new BorderLayout ());

      morphPanel = new MorphPanel ();
      NebraskaMorph.morphPanel = morphPanel;
      add ("Center", morphPanel);

      Panel control = new Panel ();
      control.setLayout (new FlowLayout ());

      newMorph = new Button ("New morph");

      colorEntryRed   = new TextField ("0", 3);
      colorEntryGreen = new TextField ("0", 3);
      colorEntryBlue  = new TextField ("128", 3);

      xSizeEntry = new TextField ("50", 4);
      ySizeEntry = new TextField ("50", 4);

      control.add (newMorph);

      control.add (new Label ("Color:"));
      control.add (colorEntryRed);
      control.add (colorEntryGreen);
      control.add (colorEntryBlue);

      control.add (new Label ("  x size:"));
      control.add (xSizeEntry);

      control.add (new Label ("  y size:"));
      control.add (ySizeEntry);

      add ("South", control);
   }
}

class NebraskaMorph extends Morph
{
   private String id;

   public String toString ()
   {
      return id;
   }

   public NebraskaMorph (String id)
   {
      super ();
      this.id = id;
      objects.put (id, this);
   }

   public NebraskaMorph (int x, int y, int z, int w, int h, Color bg, Color fg)
   {
      super (x, y, z, w, h, bg, fg);

      ++nextIDNum;
      this.id = "" + address + ":" + nextIDNum;

      objects.put ("" + this, this);

      Vector updates = new Vector ();
      updates.addElement (new NebraskaUpdate ("X", "" + GetX ()));
      updates.addElement (new NebraskaUpdate ("Y", "" + GetY ()));
      updates.addElement (new NebraskaUpdate ("Z", "" + GetZ ()));
      updates.addElement (new NebraskaUpdate ("Width", "" + GetWidth ()));
      updates.addElement (new NebraskaUpdate ("Height", "" + GetHeight ()));
      updates.addElement (new NebraskaUpdate ("Color", "" + bgColor.getRGB ()));
      SendUpdates (updates);
   }

   public void MoveTo (int x, int y)
   {
      super.MoveTo (x, y);
      Vector updates = new Vector ();
      updates.addElement (new NebraskaUpdate ("X", "" + GetX ()));
      updates.addElement (new NebraskaUpdate ("Y", "" + GetY ()));
      SendUpdates (updates);
   }
   public void QuietMoveTo (int x, int y)
   {
      super.MoveTo (x, y);
   }

   public void SetZ (int z)
   {
      super.SetZ (z);
      SendUpdate ("Z", "" + GetZ ());
   }
   public void QuietSetZ (int z)
   {
      super.SetZ (z);
   }

   private void SendUpdate (String key, String value)
   {
      Vector v = new Vector ();
      v.addElement (new NebraskaUpdate (key, value));
      SendUpdates (v);
   }

   private void SendUpdates (Vector updates)
   {
      String message = "Object \"" + this + "\"\n";
      Enumeration e = updates.elements ();
      while (e.hasMoreElements ())
         {
         NebraskaUpdate u = (NebraskaUpdate) e.nextElement ();
         message = message + u.key + " " + u.value + "\n";
         }

      if (network != null)
         {
         byte[] data = new byte[message.length () + 2];
         message.getBytes (0, message.length (), data, 2);
         data[0] = (byte) ((message.length () >> 8) & 0xff);
         data[1] = (byte) (message.length () & 0xff);

         NetworkPacket p = new NetworkPacket ();
         p.sourceAddr = address;
         p.destinationAddr = new Address ("multicast");
         p.data = data;

         network.TransmitPacket (p);
         }
   }

   public static void ProcessUpdate (String message)
   {
      StreamTokenizer msg = new StreamTokenizer (new StringBufferInputStream (message));
      try
         {
         if (msg.nextToken () != StreamTokenizer.TT_WORD)
            throw new UpdateParseError ();
         if (!msg.sval.equals ("Object"))
            throw new UpdateParseError ();

         if (msg.nextToken () != '\"')
            throw new UpdateParseError ();
         String objID = msg.sval;


         NebraskaMorph obj = (NebraskaMorph) objects.get (objID);
         if (obj == null)
            {
            obj = new NebraskaMorph (objID);
            morphPanel.AddMorph (obj);
            }
            

         int nextToken = msg.nextToken ();
         while (nextToken != StreamTokenizer.TT_EOF)
            {
            if (nextToken != StreamTokenizer.TT_WORD)
               throw new UpdateParseError ();

            if (msg.sval.equals ("X"))
               {
               if (msg.nextToken () != StreamTokenizer.TT_NUMBER)
                  throw new UpdateParseError ();
               obj.QuietMoveTo ((int) msg.nval, obj.GetY ());
               }
            else if (msg.sval.equals ("Y"))
               {
               if (msg.nextToken () != StreamTokenizer.TT_NUMBER)
                  throw new UpdateParseError ();
               obj.QuietMoveTo (obj.GetX (), (int) msg.nval);
               }
            else if (msg.sval.equals ("Z"))
               {
               if (msg.nextToken () != StreamTokenizer.TT_NUMBER)
                  throw new UpdateParseError ();
               obj.QuietSetZ ((int) msg.nval);
               obj.GetPanel ().MorphZChanged (obj);
               }

            else if (msg.sval.equals ("Width"))
               {
               if (msg.nextToken () != StreamTokenizer.TT_NUMBER)
                  throw new UpdateParseError ();
               obj.SetWidth ((int) msg.nval);
               }
            else if (msg.sval.equals ("Height"))
               {
               if (msg.nextToken () != StreamTokenizer.TT_NUMBER)
                  throw new UpdateParseError ();
               obj.SetHeight ((int) msg.nval);
               }

            else if (msg.sval.equals ("Color"))
               {
               if (msg.nextToken () != StreamTokenizer.TT_NUMBER)
                  throw new UpdateParseError ();
               obj.SetBgColor (new Color ((int) msg.nval));
               }

            nextToken = msg.nextToken ();
            }

         obj.QuietMoveTo (obj.GetX (), obj.GetY ());
         morphPanel.ForceUpdate ();
         }
      catch (UpdateParseError upe)
         {
         System.out.println ("UpdateParseError");
         }
      catch (IOException ioe)
         {
         System.out.println ("IOException");
         }
   }

   private static Hashtable objects = new Hashtable ();
   public static MorphPanel morphPanel;
   public static Network network;
   public static Address address;
   private static int nextIDNum = 0;
}

class NebraskaUpdate
{
   public NebraskaUpdate (String k, String v)
   {
      key = k;
      value = v;
   }

   public String key;
   public String value;
}

class UpdateParseError extends Exception
{
}
