package network;

import link.*;
import structure.*;

import java.util.*;

/***************************************************************** 
   The extra data carried by a network-level packet.  This 
   extends the <tt>LinkPacket</tt> structure in the <tt>link</tt>
   package, from which it inherits the field
<pre>   public byte[] data;
</pre>
   The network level adds addressing.  <tt>NetworkPacket</tt> 
   is extended by the reliability layer with sequence numbers
   and control bits.
   <p>
   Like all packets, NetworkPacket implements a toString
   for use by packet tracing.
   <p>
   While you are programming in the network level, feel free
   to add reasonable fields here.

@see LinkPacket
 *****************************************************************/
public class NetworkPacket extends LinkPacket implements StructuredObject
{
   /************************************************************
      Constructor for some default setup.  The defaults are
      <ul>
      <li> timeToLive = 10
      </ul>
    ************************************************************/
   public NetworkPacket ()
   {
      timeToLive = 10;
   }

   /************************************************************
      Make a string for packet trace messages.  Should be short
    ************************************************************/
   public String toString ()
   {
      return "(Pack:" + destinationAddr + ")";
   }

   public void FoldInStructure (StructuredStorage s)
   {
      super.FoldInStructure (s);

      sourceAddr = (Address) s.GetObject ("sourceAddr");
      destinationAddr = (Address) s.GetObject ("destinationAddr");
      timeToLive = s.GetInt ("timeToLive");

      // Your fields need to be extracted here
      linkStateMessage = (LinkStateMessage) s.GetObject ("linkStateMessage");
   }

   public void FoldOutStructure (StructuredStorage s)
   {
      super.FoldOutStructure (s);

      s.SetObject ("sourceAddr", sourceAddr);
      s.SetObject ("destinationAddr", destinationAddr);
      s.SetInt ("timeToLive", timeToLive);

      // Your fields need to be stored here
      s.SetObject ("linkStateMessage", linkStateMessage);
   }

   /************************************************************
     Address of the packet's sender
    ************************************************************/
   public Address sourceAddr;

   /************************************************************
     Address of the packet's intended recipient
    ************************************************************/
   public Address destinationAddr;

   /************************************************************
      Maximum number of hops before this packet dies
    ************************************************************/
   public int timeToLive;

   public LinkStateMessage linkStateMessage;
}
