package structure;

import java.util.*;

public class StructuredStorage
{
   public static final int INT = 1;
   public static final int DOUBLE = 2;
   public static final int BOOLEAN = 3;
   public static final int STRING = 4;
   public static final int BLOB = 5;
   public static final int RECURSIVE_OBJECT = 6;
   public static final int VECTOR = 7;

   public StructuredStorage ()
   {
      representation = new Hashtable ();
   }

   public StructuredStorage (byte[] data)
   {
      this ();
      ConvertFromByteArray (data);
   }

   private Hashtable representation;
   private String className;

   public String toString ()
   {
      String retval = "{ ";
      Enumeration e = representation.keys ();
      while (e.hasMoreElements ())
         {
         Object k = e.nextElement ();
         Object v = representation.get (k);
         retval = retval + " (" + k + " -> " + v + ")  ";
         }
      return retval + " }";
   }

   public StructuredObject ConvertToObject ()
   {
      StructuredObject obj = null;

      try
         {
         obj = (StructuredObject) Class.forName (className).newInstance ();
         }
      catch (ClassNotFoundException e)
         {
         throw new NoValueException ("Could not find class '" + className + "'");
         }
      catch (InstantiationException ie)
         {
         throw new NoValueException ("Could not instantiate class '" + className + "'");
         }
      catch (IllegalAccessException iae)
         {
         throw new NoValueException ("Could not call zero-arg constructor for '" + className + "'");
         }

      obj.FoldInStructure (this);
      return obj;
   }

   public void ConvertFromObject (StructuredObject obj)
   {
      representation = new Hashtable ();
      className = obj.getClass ().getName ();
      obj.FoldOutStructure (this);
   }

   public byte[] ConvertToByteArray ()
   {
      byte[] num;

      Vector chunks = new Vector ();

      if (className == null)
         System.out.println ("no className!");

      num = new byte[4];
      num[0] = (byte) ((className.length () >> 24) & 0xff);
      num[1] = (byte) ((className.length () >> 16) & 0xff);
      num[2] = (byte) ((className.length () >>  8) & 0xff);
      num[3] = (byte) ((className.length () >>  0) & 0xff);
      chunks.addElement (num);

      byte[] nameAsBytes = new byte[className.length ()];
      className.getBytes (0, className.length (), nameAsBytes, 0);

      chunks.addElement (nameAsBytes);

      Enumeration e = representation.keys ();
      while (e.hasMoreElements ())
         {
         String name = (String) e.nextElement ();
         GenericDataPair pair = (GenericDataPair) representation.get (name);

         if (pair == null)
            System.out.println ("no pair for " + name);

         num = new byte[4];
         num[0] = (byte) ((name.length () >> 24) & 0xff);
         num[1] = (byte) ((name.length () >> 16) & 0xff);
         num[2] = (byte) ((name.length () >>  8) & 0xff);
         num[3] = (byte) ((name.length () >>  0) & 0xff);
         chunks.addElement (num);

         nameAsBytes = new byte[name.length ()];
         name.getBytes (0, name.length (), nameAsBytes, 0);

         chunks.addElement (nameAsBytes);
         byte[] rep = pair.ConvertToByteArray ();
         chunks.addElement (rep);
         }

      int size = 0;
      e = chunks.elements ();
      while (e.hasMoreElements ())
         {
         byte[] b = (byte[]) e.nextElement ();
         size += b.length;
         }

      byte[] retval = new byte[size+4];

      size = 0;
      e = chunks.elements ();
      while (e.hasMoreElements ())
         {
         byte[] b = (byte[]) e.nextElement ();

         for (int i=0; i<b.length; ++i)
            retval[i+size] = b[i];

         size += b.length;
         }
      
      retval[size + 0] = (byte) 0;
      retval[size + 1] = (byte) 0;
      retval[size + 2] = (byte) 0;
      retval[size + 3] = (byte) 0;

      return retval;
   }

   public static int ReadNumber (byte[] data, int offset)
   {
      int retval = 0;
      retval |= (data[offset+0] & 0xff) << 24;
      retval |= (data[offset+1] & 0xff) << 16;
      retval |= (data[offset+2] & 0xff) << 8;
      retval |= (data[offset+3] & 0xff) << 0;

      if (retval < 0)
         System.out.println ("OUCH");

      return retval;
   }

   public void ConvertFromByteArray (byte[] data)
   {
      int cnLen = ReadNumber (data, 0);
      className = new String (data, 0, 4, cnLen);

      representation = new Hashtable ();
      int offset = 4 + cnLen;

      for (;;)
         {
         int fnLen = ReadNumber (data, offset);

         if (fnLen == 0)
            break;

         String name = new String (data, 0, offset+4, fnLen);
         offset += (4 + fnLen);

         int dataLen = ReadNumber (data, offset+1);

         byte[] fieldData = new byte[5+dataLen];
         for (int i=0; i<(5+dataLen); ++i)
            fieldData[i] = data[offset+i];
         offset += (5 + dataLen);

         GenericDataPair pair = null;

         switch ((int) fieldData[0])
            {
            case INT:
               pair = new IntDataPair (fieldData);
               break;
            case DOUBLE:
               pair = new DoubleDataPair (fieldData);
               break;
            case BOOLEAN:
               pair = new BooleanDataPair (fieldData);
               break;
            case STRING:
               pair = new StringDataPair (fieldData);
               break;
            case BLOB:
               pair = new BLOBDataPair (fieldData);
               break;
            case RECURSIVE_OBJECT:
               pair = new RecursiveObjectDataPair (fieldData);
               break;
            case VECTOR:
               pair = new VectorDataPair (fieldData);
               break;
            default:
               System.out.println ("Unknown type code: " + (int) fieldData[0]);
            }

         representation.put (name, pair);
         }
   }

   public void SetInt (String name, int val)
   {
      IntDataPair pair = new IntDataPair ();
      pair.data = val;
      representation.put (name, pair);
   }

   public void SetDouble (String name, double val)
   {
      DoubleDataPair pair = new DoubleDataPair ();
      pair.data = val;
      representation.put (name, pair);
   }

   public void SetBoolean (String name, boolean val)
   {
      BooleanDataPair pair = new BooleanDataPair ();
      pair.data = val;
      representation.put (name, pair);
   }

   public void SetString (String name, String val)
   {
      StringDataPair pair = new StringDataPair ();
      pair.data = val;
      representation.put (name, pair);
   }

   public void SetObject (String name, StructuredObject val)
   {
      RecursiveObjectDataPair pair = new RecursiveObjectDataPair ();

      if (val != null)
         {
         StructuredStorage s = new StructuredStorage ();
         s.ConvertFromObject (val);
         pair.data = s;
         }
      else
         pair.data = null;

      representation.put (name, pair);
   }

   public void SetBLOB (String name, byte[] val)
   {
      BLOBDataPair pair = new BLOBDataPair ();
      pair.data = val;
      representation.put (name, pair);
   }

   public void SetVector (String name, Vector val)
   {
      VectorDataPair pair = new VectorDataPair ();
      pair.data = new Vector ();
      Enumeration e = val.elements ();
      while (e.hasMoreElements ())
         {
         StructuredObject o = (StructuredObject) e.nextElement ();
         StructuredStorage s;

         if (o == null)
            s = null;
         else
            {
            s = new StructuredStorage ();
            s.ConvertFromObject (o);
            }

         pair.data.addElement (s);
         }
      representation.put (name, pair);
   }

   public int GetInt (String name) throws NoValueException
   {
      IntDataPair pair = (IntDataPair) representation.get (name);
      if (pair == null)
         throw new NoValueException ("No value in representation for name '" + name + "'");
      if (!(pair instanceof IntDataPair))
         throw new NoValueException ("Value at '" + name + "' is of incorrect type");

      return pair.data;
   }

   public double GetDouble (String name) throws NoValueException
   {
      DoubleDataPair pair = (DoubleDataPair) representation.get (name);
      if (pair == null)
         throw new NoValueException ("No value in representation for name '" + name + "'");
      if (!(pair instanceof DoubleDataPair))
         throw new NoValueException ("Value at '" + name + "' is of incorrect type");

      return pair.data;
   }

   public boolean GetBoolean (String name) throws NoValueException
   {
      BooleanDataPair pair = (BooleanDataPair) representation.get (name);
      if (pair == null)
         throw new NoValueException ("No value in representation for name '" + name + "'");
      if (!(pair instanceof BooleanDataPair))
         throw new NoValueException ("Value at '" + name + "' is of incorrect type");

      return pair.data;
   }

   public String GetString (String name) throws NoValueException
   {
      StringDataPair pair = (StringDataPair) representation.get (name);
      if (pair == null)
         throw new NoValueException ("No value in representation for name '" + name + "'");
      if (!(pair instanceof StringDataPair))
         throw new NoValueException ("Value at '" + name + "' is of incorrect type");

      return pair.data;
   }

   public StructuredObject GetObject (String name) throws NoValueException
   {
      RecursiveObjectDataPair pair = (RecursiveObjectDataPair) representation.get (name);
      if (pair == null)
         throw new NoValueException ("No value in representation for name '" + name + "'");
      if (!(pair instanceof RecursiveObjectDataPair))
         throw new NoValueException ("Value at '" + name + "' is of incorrect type");

      StructuredStorage s = pair.data;
      if (s != null)
         return s.ConvertToObject ();
      else
         return null;
   }

   public byte[] GetBLOB (String name) throws NoValueException
   {
      BLOBDataPair pair = (BLOBDataPair) representation.get (name);
      if (pair == null)
         throw new NoValueException ("No value in representation for name '" + name + "'");
      if (!(pair instanceof BLOBDataPair))
         throw new NoValueException ("Value at '" + name + "' is of incorrect type");

      return pair.data;
   }

   public Vector GetVector (String name) throws NoValueException
   {
      VectorDataPair pair = (VectorDataPair) representation.get (name);
      if (pair == null)
         throw new NoValueException ("No value in representation for name '" + name + "'");
      if (!(pair instanceof VectorDataPair))
         throw new NoValueException ("Value at '" + name + "' is of incorrect type");

      Vector retval = new Vector ();
      Enumeration e = pair.data.elements ();
      while (e.hasMoreElements ())
         {
         StructuredStorage s = (StructuredStorage) e.nextElement ();
         if (s != null)
            {
            StructuredObject o = s.ConvertToObject ();
            retval.addElement (o); 
            }
         else
            retval.addElement (null);
         }

      return retval;
   }
}

abstract class GenericDataPair
{
   public abstract byte[] ConvertToByteArray ();
}

class IntDataPair extends GenericDataPair
{
   private static int code = StructuredStorage.INT;

   public int data;

   public IntDataPair ()
   {
   }

   public IntDataPair (byte[] asBytes)
   {
      data = 0;
      data = StructuredStorage.ReadNumber (asBytes, 5);
   }

   public byte[] ConvertToByteArray ()
   {
      byte[] retval = new byte[9];
      retval[0] = (byte) code;

      retval[1] = (byte) 0;
      retval[2] = (byte) 0;
      retval[3] = (byte) 0;
      retval[4] = (byte) 4;

      retval[5] = (byte) ((data >> 24) & 0xff);
      retval[6] = (byte) ((data >> 16) & 0xff);
      retval[7] = (byte) ((data >>  8) & 0xff);
      retval[8] = (byte) ((data >>  0) & 0xff);

      return retval;      
   }
}

class DoubleDataPair extends GenericDataPair
{
   private static int code = StructuredStorage.DOUBLE;

   public double data;

   public DoubleDataPair ()
   {
      data = 0.0;
   }

   public DoubleDataPair (byte[] asBytes)
   {
      data = 0.1;
      long asLong = 0;
      asLong |= (asBytes[5] & 0xff) << 56;
      asLong |= (asBytes[6] & 0xff) << 48;
      asLong |= (asBytes[7] & 0xff) << 40;
      asLong |= (asBytes[8] & 0xff) << 32;
      asLong |= (asBytes[9] & 0xff) << 24;
      asLong |= (asBytes[10] & 0xff) << 16;
      asLong |= (asBytes[11] & 0xff) << 8;
      asLong |= (asBytes[12] & 0xff) << 0;

      data = Double.longBitsToDouble (asLong);
   }

   public byte[] ConvertToByteArray ()
   {
      byte[] retval = new byte[13];
      retval[0] = (byte) code;

      retval[1] = (byte) 0;
      retval[2] = (byte) 0;
      retval[3] = (byte) 0;
      retval[4] = (byte) 8;

      long asLong = Double.doubleToLongBits (data);

      retval[5] = (byte) ((asLong >> 56) & 0xff);
      retval[6] = (byte) ((asLong >> 48) & 0xff);
      retval[7] = (byte) ((asLong >> 40) & 0xff);
      retval[8] = (byte) ((asLong >> 32) & 0xff);
      retval[9] = (byte) ((asLong >> 24) & 0xff);
      retval[10] = (byte) ((asLong >> 16) & 0xff);
      retval[11] = (byte) ((asLong >>  8) & 0xff);
      retval[12] = (byte) ((asLong >>  0) & 0xff);

      return retval;
   }
}

class BooleanDataPair extends GenericDataPair
{
   private static int code = StructuredStorage.BOOLEAN;

   public boolean data;

   public BooleanDataPair ()
   {
   }

   public BooleanDataPair (byte[] asBytes)
   {
      data = (asBytes[5] == 1);
   }

   public byte[] ConvertToByteArray ()
   {
      byte[] retval = new byte[6];
      retval[0] = (byte) code;

      retval[1] = (byte) 0;
      retval[2] = (byte) 0;
      retval[3] = (byte) 0;
      retval[4] = (byte) 1;

      retval[5] = (byte) (data ? 1 : 0);
      return retval;
   }
}

class StringDataPair extends GenericDataPair
{
   private static int code = StructuredStorage.STRING;

   public String data;

   public StringDataPair ()
   {
   }

   public StringDataPair (byte[] asBytes)
   {
      int length = 0;
      length = StructuredStorage.ReadNumber (asBytes, 1);

      if (length == 1)
         data = null;
      else
         data = new String (asBytes, 0, 5, length);
   }

   public byte[] ConvertToByteArray ()
   {
      if (data == null)
         {
         byte[] retval = new byte[6];

         retval[0] = (byte) code;
         retval[1] = (byte) 0;
         retval[2] = (byte) 0;
         retval[3] = (byte) 0;
         retval[4] = (byte) 1;
         retval[5] = (byte) 0;

         return retval;
         }
      else
         {
         byte[] retval = new byte[5 + data.length ()];
         retval[0] = (byte) code;

         retval[1] = (byte) ((data.length () >> 24) & 0xff);
         retval[2] = (byte) ((data.length () >> 16) & 0xff);
         retval[3] = (byte) ((data.length () >>  8) & 0xff);
         retval[4] = (byte) ((data.length () >>  0) & 0xff);

         data.getBytes (0, data.length (), retval, 5);

         return retval;
         }
   }
}

class BLOBDataPair extends GenericDataPair
{
   private static int code = StructuredStorage.BLOB;

   public byte[] data;

   public BLOBDataPair ()
   {
   }

   public BLOBDataPair (byte[] asBytes)
   {
      int length = StructuredStorage.ReadNumber (asBytes, 1);

      data = new byte[length];
      for (int i=0; i<length; ++i)
         data[i] = asBytes[i+5];
   }

   public byte[] ConvertToByteArray ()
   {
      byte[] retval = new byte[5 + data.length];

      retval[0] = (byte) code;

      retval[1] = (byte) ((data.length >> 24) & 0xff);
      retval[2] = (byte) ((data.length >> 16) & 0xff);
      retval[3] = (byte) ((data.length >>  8) & 0xff);
      retval[4] = (byte) ((data.length >>  0) & 0xff);

      for (int i=0; i<data.length; ++i)
         retval[i+5] = data[i];

      return retval;
   }
}

class RecursiveObjectDataPair extends GenericDataPair
{
   private static int code = StructuredStorage.RECURSIVE_OBJECT;

   public StructuredStorage data;

   public RecursiveObjectDataPair ()
   {
   }

   public RecursiveObjectDataPair (byte[] asBytes)
   {
      int length = StructuredStorage.ReadNumber (asBytes, 1);

      byte[] d = new byte[length];
      for (int i=0; i<length; ++i)
         d[i] = asBytes[i+5];

      if (length == 1)
         {
         data = null;
         }
      else
         {
         data = new StructuredStorage ();
         data.ConvertFromByteArray (d);
         }
   }

   public byte[] ConvertToByteArray ()
   {
      byte[] contents;

      if (data != null)
         contents = data.ConvertToByteArray ();
      else
         {
         contents = new byte[1];
         contents[0] = 0;
         }

      byte[] retval = new byte[5 + contents.length];

      retval[0] = (byte) code;

      retval[1] = (byte) ((contents.length >> 24) & 0xff);
      retval[2] = (byte) ((contents.length >> 16) & 0xff);
      retval[3] = (byte) ((contents.length >>  8) & 0xff);
      retval[4] = (byte) ((contents.length >>  0) & 0xff);

      for (int i=0; i<contents.length; ++i)
         retval[i+5] = contents[i];

      return retval;
   }
}

class VectorDataPair extends GenericDataPair
{
   private static int code = StructuredStorage.VECTOR;

   public Vector data;

   public VectorDataPair ()
   {
      data = new Vector ();
   }

   public VectorDataPair (byte[] asBytes)
   {
      data = new Vector ();

      int offset = 5;

      while (offset < asBytes.length)
         {
         int dataLen = StructuredStorage.ReadNumber (asBytes, offset);

         byte[] fieldData = new byte[dataLen];
         for (int i=0; i<dataLen; ++i)
            fieldData[i] = asBytes[offset+4+i];
         offset += (4 + dataLen);

         StructuredStorage s;

         if (dataLen == 1)
            {
            s = null;
            }
         else
            {
            s = new StructuredStorage ();
            s.ConvertFromByteArray (fieldData);
            }

         data.addElement (s);
         }
   }

   public byte[] ConvertToByteArray ()
   {
      Vector chunks = new Vector ();
     
      Enumeration e = data.elements ();
      while (e.hasMoreElements ())
         {
         StructuredStorage s = (StructuredStorage) e.nextElement ();

         byte[] contents;

         if (s != null)
            contents = s.ConvertToByteArray ();
         else
            {
            contents = new byte[1];
            contents[0] = 0;
            }

         byte[] header = new byte[4];

         header[0] = (byte) ((contents.length >> 24) & 0xff);
         header[1] = (byte) ((contents.length >> 16) & 0xff);
         header[2] = (byte) ((contents.length >>  8) & 0xff);
         header[3] = (byte) ((contents.length >>  0) & 0xff);

         chunks.addElement (header);
         chunks.addElement (contents);
         }

      int size = 0;
      e = chunks.elements ();
      while (e.hasMoreElements ())
         {
         byte[] b = (byte[]) e.nextElement ();
         size += b.length;
         }

      byte[] retval = new byte[size+5];
      retval[0] = (byte) code;

      size = 0;
      e = chunks.elements ();
      while (e.hasMoreElements ())
         {
         byte[] b = (byte[]) e.nextElement ();

         for (int i=0; i<b.length; ++i)
            retval[i+size+5] = b[i];

         size += b.length;
         }

      retval[1] = (byte) ((size >> 24) & 0xff);
      retval[2] = (byte) ((size >> 16) & 0xff);
      retval[3] = (byte) ((size >>  8) & 0xff);
      retval[4] = (byte) ((size >>  0) & 0xff);

      return retval;
   }
}

class NoValueException extends RuntimeException
{
   public NoValueException (String s)
   {
      super (s);
   }
}

